﻿using System;
using System.Drawing;
using System.Drawing.Drawing2D;


namespace Pololu
{
    public class Channel
    {
        // This class maintains the data plot for a single channel drawn on a transparent bitmap.
        // Two bitmaps are used for drawing streaming (long-period) data, but only one is ever active at
        // a given time (when space on the current bitmap runs out, it is copied to the other bitmap
        // at a shifted location so there is room to continue drawing the data).
        // The bitmaps in this object are scaled horizontally so the number of points to draw fits
        // exactly within the width of the bitmap.  There is no vertical scaling.

        private Bitmap[] bmp = new Bitmap[2] {null, null};
        private Graphics[] bmpGraphics = new Graphics[2] {null, null};
        private int bmpIdx = 0;

        private int pointsToDrawInternal = 1000;   // default value to prevent possible divide-by-zero error
        private int widthInternal;
        private float voltsPerDivInternal = 1;
        private int scrollPos = 0;

        Point lastPoint = new Point(-1, -1);

        int numPensInternal = 5;
        float penWidthInternal = 2;
        Pen[] pen;
        Point[][] prevPath;
        Point[] curPath = null;


        public Channel(Color penColor, int width)
        {
            for (int i = 0; i < 2; i++)
            {
                bmp[i] = new Bitmap(width * 2, 520);        // Note: make the height a bigger than the minimum 256 pixels we need;
                                                            // this helps avoid rendering problems that makes pixels at 0 and 255 disappear
                                                            // when the bitmap is scaled and drawn to the screen.  We use 520 so there is
                                                            // room to draw the waveform when using 0.5 V/div.
                bmpGraphics[i] = Graphics.FromImage(bmp[i]);
                bmpGraphics[i].SmoothingMode = SmoothingMode.AntiAlias;
                bmpGraphics[i].Clear(Color.Transparent);
            }

            pen = new Pen[NumPens];
            prevPath = new Point[NumPens - 1][];

            Width = width;          // this line *MUST* come after bmpGraphics[i] has been initialized
            PenColor = penColor;    // initializes all of the channel pens (Width must be set before this line is called)
        }


        ~Channel()
        {
            this.Dispose();
        }


        public void Dispose()
        {
            for (int i = 0; i < 2; i++)
            {
                if (bmp[i] != null)
                    bmp[i].Dispose();
                if (bmpGraphics[i] != null)
                    bmpGraphics[i].Dispose();
            }
            for (int i = 0; i < NumPens; i++)
                pen[i].Dispose();
        }


        public Color PenColor
        {
            get { return pen[0] == null ? Color.Transparent : pen[0].Color; }
            set
            {
                for (int i = 0; i < NumPens; i++)
                {
                    Color c = Color.FromArgb(255 * (NumPens - i) / NumPens, value);
                    if (pen[i] == null)
                        pen[i] = new Pen(c, PenWidth);
                    else
                        pen[i].Color = c;
                }
                ApplyTransform(0);
            }
        }


        public Pen ChannelPen
        {
            get { return pen[0]; }
        }


        public int Width
        {
            get { return widthInternal; }
            set
            {
                widthInternal = value;
                UpdateTransforms();
                // To Do: recalculate the proper quantities if this changes
            }
        }


        public float PenWidth
        {
            get { return penWidthInternal; }
            set
            {
                penWidthInternal = value;
                for (int i = 0; i < NumPens; i++)
                {
                    if (pen[i] == null)
                        pen[i] = new Pen(Color.Transparent, penWidthInternal);
                    else
                        pen[i].Width = penWidthInternal;
                }
                ApplyTransform(0);
            }
        }


        public int PointsToDraw
        {
            get { return pointsToDrawInternal; }
            set
            {
                pointsToDrawInternal = value;
                if (pointsToDrawInternal > Width)
                    PenWidth = 1;
                else
                    PenWidth = 2;
                UpdateTransforms();      // this method will reset various fields and clear the bitmaps
            }
        }


        public float VoltsPerDivision
        {
            get { return voltsPerDivInternal; }
            set
            {
                voltsPerDivInternal = value;
            }
        }


        public int NumPens
        {
            get { return numPensInternal; }
            set
            {
                Color c = PenColor;
                for (int i = 0; i < numPensInternal; i++)
                    pen[i].Dispose();
                numPensInternal = value + 1;
                if (numPensInternal <= 1)
                    numPensInternal = 2;
                pen = new Pen[numPensInternal];
                PenColor = c;
                prevPath = new Point[numPensInternal - 1][];
                for (int i = 0; i < prevPath.Length; i++)
                    prevPath[i] = null;
            }
        }


        public int[] MaxYAndMinY
        {
            get
            {
                int[] maxAndMin = new int[2] { -1, -1 };
                if (curPath == null)
                    return maxAndMin;
                maxAndMin[0] = curPath[0].Y;
                maxAndMin[1] = curPath[0].Y;
                for (int i = 1; i < curPath.Length; i++)
                {
                    if (maxAndMin[0] < curPath[i].Y)
                        maxAndMin[0] = curPath[i].Y;
                    if (maxAndMin[1] > curPath[i].Y)
                        maxAndMin[1] = curPath[i].Y;
                }
                return maxAndMin;
            }
        }

        public Matrix Transform
        {
            get { return bmpGraphics[0].Transform; }
        }


        public void Reset()
        {
            bmpIdx = 0;
            for (int i = 0; i < 2; i++)
                bmpGraphics[i].Clear(Color.Transparent);

            scrollPos = 0;
            lastPoint = new Point(-1, -1);

            for (int i = 0; i < prevPath.Length; i++)
                prevPath[i] = null;
            curPath = null;
        }


        private void ApplyTransform(int i)
        {
            bmpGraphics[i].ResetTransform();
            bmpGraphics[i].TranslateTransform(0, 1);
            bmpGraphics[i].ScaleTransform((float)Width / PointsToDraw, 1);
            for (int j = 0; j < NumPens; j++)
            {
                if (pen[j] == null)
                    continue;
                pen[j].ResetTransform();
                pen[j].ScaleTransform((float)PointsToDraw / Width, 1);
            }
        }


        private void UpdateTransforms()
        {
            for (int i = 0; i < 2; i++)
            {
                ApplyTransform(i);
            }
            Reset();
        }


        private void SwitchBitmaps()
        {
            bmpIdx = (bmpIdx + 1) % 2;
            bmpGraphics[bmpIdx].Clear(Color.Transparent);
            scrollPos -= PointsToDraw;
            lastPoint.X -= PointsToDraw;
            bmpGraphics[bmpIdx].ResetTransform();
            //bmpGraphics[bmpIdx].DrawImage(bmp[(bmpIdx + 1) % 2], -PointsToDraw, 0);
            bmpGraphics[bmpIdx].DrawImage(bmp[(bmpIdx + 1) % 2], -Width, 0);
            ApplyTransform(bmpIdx);
        }


        public int UpdateScrollingData(Point[] points)
        {
            GraphicsPath path = new GraphicsPath();

            if (scrollPos + points.Length > 2 * PointsToDraw)
            {
                SwitchBitmaps();
            }

            for (int i = 0; i < points.Length; i++)
            {
                points[i].X += scrollPos;
                points[i].Y = (int)Math.Round((float)points[i].Y / VoltsPerDivision);
            }

            if (lastPoint.Y >= 0)                       // if the last point of the last data set is valid
                path.AddLine(lastPoint, points[0]);     // use it as the starting point of this data set
            path.AddLines(points);
            lastPoint = points[points.Length - 1];      
            bmpGraphics[bmpIdx].DrawPath(ChannelPen, path);     // draw the path to the currently active bitmap
            scrollPos += points.Length;                 // update the X coordinate of our current drawing position

            return scrollPos;                           // return this coordinate
        }


        public void UpdateFullScreenData(Point[] points, bool persistence)
        {
            GraphicsPath path = new GraphicsPath();

            bmpGraphics[bmpIdx].Clear(Color.Transparent);

            if (points != null)
            {
                for (int i = 0; i < points.Length; i++)
                {
                    points[i].Y = (int)Math.Round((float)points[i].Y / VoltsPerDivision);
                }
                prevPath[0] = curPath;
                curPath = (Point[])points.Clone();
            }
            else
                prevPath[0] = null;

            for (int i = NumPens - 1; i > 0; i--)
            {
                if (prevPath[i - 1] != null && persistence)
                {
                    path.AddLines(prevPath[i - 1]);
                    bmpGraphics[bmpIdx].DrawPath(pen[i], path);
                    path.Reset();
                }
                if (i >= 2)
                    prevPath[i - 1] = prevPath[i - 2];
            }

            if (curPath != null)
            {
                path.AddLines(curPath);
                bmpGraphics[bmpIdx].DrawPath(pen[0], path);
                scrollPos = PointsToDraw;
            }

        }


        public void Draw(Graphics g, int verticalPos)
        {
            g.DrawImage(bmp[bmpIdx], Width - (float)scrollPos * Width / PointsToDraw, verticalPos);//, bmp[bmpIdx].Width, (float)bmp[bmpIdx].Height / VoltsPerDivision);
        }
    }
}
