﻿
// TO DO: save settings
// TO DO: add triggering
// TO DO: allow for window resizing, add more vertical divisions
// TO DO: add measurement taking (curve voltage at cursor, dV and dT)

using System;
using System.Drawing;
using System.Windows.Forms;
using System.Drawing.Drawing2D;
using System.Text;
using System.ComponentModel;


namespace Pololu
{
    public partial class Form1 : Form
    {
        /// <summary>
        /// Determines how many readings we hold in memory at once.
        /// </summary>
        const int memorySize = 120000;

        /// <summary>
        /// The number of points per pixel to draw when displaying the entire contents of the buffer in
        /// the allData panel.  Drawing too many points can get too processor intensive; drawing too
        /// few points won't look as good.
        /// </summary>
        const int allDataPointsPerPixel = 4;

        /// <summary>
        /// The number of pixels of margin for the top of the graph.
        /// </summary>
        const int topGraphMargin = 10;

        /// <summary>
        /// The number of pixels of margin for the bottom of the graph.
        /// </summary>
        const int bottomGraphMargin = 10;

        /// <summary>
        /// The number of pixels of margin for the top of the graph.
        /// </summary>
        const int topAllDataMargin = 2;

        /// <summary>
        /// The number of pixels of margin for the bottom of the graph.
        /// </summary>
        const int bottomAllDataMargin = 2;

        /// <summary>
        /// Number of horizontal and vertical divisions to draw in the graph panel.
        /// The code is written in a way that editing these values should achieve the
        /// desired changes without having to make any additional alterations to the code.
        /// Future improvements to this program could involve making these settable
        /// by form controls while the program is running.
        /// </summary>
        const int numHorizDivisions = 10;
        const int numVertDivisions = 6;

        /// <summary>
        /// If the data on the screen spans a time period greater than
        /// streamPeriodThreshold milliseconds, the data will be streamed rather
        /// than drawn as a single full-screen update.  Performing full-screen
        /// updates with a lot of data makes the updates annoyingly infrequent and
        /// choppy.  Streaming small amounts of data isn't very useful since features
        /// scroll by more quickly than they can be perceived.
        /// </summary>
        const int streamPeriodThreshold = 200;



        /// <summary>
        /// An object representing the slo-scope we are connected to.
        /// </summary>
        Sloscope sloscope;

        /// <summary>
        /// Objects that help us draw the channel data on the graph panel.
        /// </summary>
        Channel channelA;
        Channel channelB;

        /// <summary>
        /// Flag that is true when the SLO-scope is enabled and streaming data.
        /// </summary>
        bool running = false;

        /// <summary>
        /// The number of readings to display on the screen at once.
        /// </summary>
        int graphHorizontalSize;

        /// <summary>
        /// The readings that we are holding in memory.
        /// </summary>
        RawData data;

        /// <summary>
        /// The queue index of the most recent data that has been drawn to the screen.
        /// A value of (data.Size-1) means that the most recently drawn data is also the
        /// most recently received data.  This variable is used to maintain the position
        /// of the purple graph panel indicator rectangle in the allData panel.
        /// </summary>
        int mostRecentlyDrawnBufferPosition = 0;

        /// <summary>
        /// The number of readings we have received since we last updated the screen with
        /// the current data.  This variable is only used when the SLO-scope is running
        /// and it lets us figure out when we have enough data to update the full graph
        /// panel.
        /// </summary>
        int pointsSinceUpdate = 0;

        /// <summary>
        /// This value lets us convert between the counts returned by the SLO-scope and
        /// volts.  It is initialized to 1000 / GetCountsPerMillivolt() in the form's load
        /// event.  See the comments to the GetCountsPerMillivolt() method for more
        /// information.  Note that it is important to set this variable early as
        /// many other parts of this program rely on it.
        /// </summary>
        double countsPerVolt = 0;

        /// <summary>
        /// Arrays of points that store the data to be drawn in the allData panel.
        /// </summary>
        Point[] allDataPathA;
        Point[] allDataPathB;

        /// <summary>
        /// Arrays of points that store the data to be drawn in the graph panel when
        /// in full-screen update mode.
        /// </summary>
        Point[] fullScreenA;
        Point[] fullScreenB;

        /// <summary>
        /// Vertical positions in graph panel space of the 0V points of channels A and B.
        /// </summary>
        int vertPosA = 0;
        int vertPosB = 0;

        /// <summary>
        /// Paths that define the triangular scrollbars used to set the 0V points of
        /// channels A and B.
        /// </summary>
        GraphicsPath  scrollbarAPath;
        GraphicsPath  scrollbarBPath;

        /// <summary>
        /// Flags triggered by and used by events to let the program know what state its in
        /// and associated data that needs to be stored to maintain these states
        /// </summary>
        bool triggerAdjustIsActive = false;     // true if we are dragging the trigger level bar around
        bool scrollbarAIsActive = false;        // true if we are dragging the channel A vertical position scroll bar around
        bool scrollbarBIsActive = false;        // true if we are dragging the channel B vertical position scroll bar around 
        bool allDataPanIsActive = false;        // true if we are dragging the purple graph panel indicator rectangle around the allData panel
        bool graphDataPanIsActive = false;      // true if we are dragging the mouse around in the graph panel to scroll through the data
        int graphDataPanStartX;                 // used when calculating the horizontal scroll position of graph panel data
        bool graphDataAnalysisActive = false;   // true if we are drawing marks on the data where the data at the x coordinate of the cursor
        Point analysisPoint = new Point(-1, -1);    // the coordinate of the analysis point in screen coordinates ((-1, -1) means "no point")
        bool reset = false;                     // true if we need to reset the data the next time the SLO-scope starts running



        public Form1()
        {
            InitializeComponent();

            // make sure the window cannot be resized
            this.MaximumSize = this.Size;
            this.MinimumSize = this.Size;
            this.MaximizedBounds = new Rectangle(0, 0, this.Size.Width, this.Size.Height);
        }


        private void Form1_Load(object sender, EventArgs e)
        {
            // Add the current version numbers to the Window's title bar
            if (System.Deployment.Application.ApplicationDeployment.IsNetworkDeployed)
                this.Text += " (" + System.Deployment.Application.ApplicationDeployment.CurrentDeployment.CurrentVersion + ")";

            // Attempt to connect to the SLO-scope
            try
            {
                sloscope = Sloscope.connect();

                // a number of later calculations depend on this countsPerVolt, so set this variable as soon as the sloscope connects
                countsPerVolt = 1000 / GetMillivoltsPerCount(50);
            }
            catch (Exception exception)
            {
                displayException(exception);
                Close();
                return;
            }

            data = new RawData(sloscope, memorySize);

            channelA = new Channel(Color.Yellow, graph.ClientSize.Width);
            colorA.BackColor = channelA.PenColor;
            channelB = new Channel(Color.Cyan, graph.ClientSize.Width);
            colorB.BackColor = channelB.PenColor;

            // graph-panel-related initializations
            graph.CreateBufferedGraphics();
            // scale graph panel's buffered graphics object appropriately (this must be called *AFTER* mvPerCount is set)
            ApplyVerticalGraphTransform(graph);
            graph.BG.Graphics.SmoothingMode = SmoothingMode.AntiAlias;
            DrawGridLines();            // draw grid lines in the graph panel

            // scrollbar-panel-related initializations
            scrollbar.CreateBufferedGraphics();
            // set scrollbar panel to have the same vertical transformations as the graph panel
            ApplyVerticalGraphTransform(scrollbar);
            scrollbar.BG.Graphics.SmoothingMode = SmoothingMode.AntiAlias;
            scrollbarAPath = new GraphicsPath();
            UpdateScrollbar(scrollbarAPath, vertPosA);
            scrollbarBPath = new GraphicsPath();
            UpdateScrollbar(scrollbarBPath, vertPosB);

            // trigger-panel-related initializations
            trigger.CreateBufferedGraphics();
            // set trigger panel to have the same vertical transformations as the graph panel
            ApplyVerticalGraphTransform(trigger);
            trigger.BG.Graphics.SmoothingMode = SmoothingMode.AntiAlias;

            // allData-panel-related initializations (transform the graphics object to make it easier to draw the data to it nicely later)
            allDataPathA = new Point[allData.ClientSize.Width * allDataPointsPerPixel];
            allDataPathB = new Point[allDataPathA.Length];

            allData.CreateBufferedGraphics();
            ApplyTransform(allData, topAllDataMargin, bottomAllDataMargin, allDataPathA.Length, 255);
            allData.BG.Graphics.SmoothingMode = SmoothingMode.AntiAlias;

            // set initial states of the input/output state selection controls
            IOA.SelectedIndex = 0;
            IOB.SelectedIndex = 0;

            triggerMode.SelectedIndex = 0;
        }

        private void Form1_FormClosed(object sender, FormClosedEventArgs e)
        {
            // dispose of our used resources and turn off the SLO-scope when the form is closed
            if (channelA != null)
                channelA.Dispose();
            if (channelB != null)
                channelB.Dispose();
            if (allData != null)
                allData.Dispose();
            if (graph != null)
                graph.Dispose();
            if (sloscope != null)
            {
                try
                {
                    sloscope.state = Sloscope.State.Off;
                }
                catch (Exception) { }
            }
        }


        private void Reset()
        {
            reset = true;   // flag that a reset of the data is needed
            if (running)    // but only perform the reset if SLO-scope is running
            {               // otherwise, retain the data until we next start running the SLO-scope
                data.Reset();
                reset = false;
                pointsSinceUpdate = 0;
            }
        }


        /// <summary>
        /// Averages the specified number of reference voltage samples from the SLO-scope, which are returned
        /// as 10-bit representations of its internal 1024 mV reference voltage.  From this, we compute and return
        /// a measure of millivolts per 8-bit reading count.  Multiplying this value by the channelA and channelB
        /// members of Sloscope.Reading objects will convert them into millivolts.
        /// </summary>
        /// <param name="samples">The number of reference voltage samples to average.</param>
        /// <returns>A measure of millivolts per 8-bit SLO-scope voltage reading count.</returns>
        private double GetMillivoltsPerCount(int samples)
        {
            double refVoltage = 0;
            for (int i = 0; i < samples; i++)
                refVoltage += sloscope.getReferenceVoltage() >> 6;  // the six least-significant bits don't matter
            refVoltage /= samples;   // this gives us the average 10-bit value that corresponds to the reference voltage of 1024 mV
            // To convert this to an 8-bit value we multiply by 255 and divide by 1023
            // then we divide 1024 mV by this quantity to get millivolts per 8-bit reading count:
            return (double)1024 * 1023 / 255 / refVoltage;
        }


        // determine the number of data points to draw in the graph panel
        private void SetGraphHorizontalSize()
        {
            int prevSize = graphHorizontalSize;

            if (allAnalogModeRB.Checked)     // 10 kHz
            {
                graphHorizontalSize = (int)(periodSelection.Value * 10 * numHorizDivisions);
            }
            else
            {
                graphHorizontalSize = (int)(periodSelection.Value * 20 * numHorizDivisions);
            }

            channelA.PointsToDraw = graphHorizontalSize;
            fullScreenA = new Point[graphHorizontalSize + 1];
            channelB.PointsToDraw = graphHorizontalSize;
            fullScreenB = new Point[graphHorizontalSize + 1];

            if (!running)
            {
                // zoom in or out on the center of the graph data (without this line, we would zoom on the right edge)
                mostRecentlyDrawnBufferPosition += graphHorizontalSize / 2 - prevSize / 2;
                MostRecentlyDrawnBoundaryCheck();   // this line should be here no matter where we zoom
            }

            Reset();
        }


        private bool InStreamingMode()
        {
            return (int)periodSelection.Value * numHorizDivisions > streamPeriodThreshold;
        }


        private void ApplyTransform(CustomPanel panel, int topMargin, int bottomMargin, float xMax, float yMax)
        {
            panel.BG.Graphics.ResetTransform();
            panel.BG.Graphics.TranslateTransform(0, panel.ClientSize.Height - bottomMargin);
            panel.BG.Graphics.ScaleTransform(panel.ClientSize.Width / xMax, 
                -(float)(((double)panel.ClientSize.Height - bottomMargin - topMargin) / yMax));
        }


        private void ApplyVerticalGraphTransform(CustomPanel panel)
        {
            ApplyTransform(panel, topGraphMargin, bottomGraphMargin, panel.ClientSize.Width, (float)countsPerVolt * numVertDivisions);
        }


        private void UpdateScrollbar(GraphicsPath scrollbarPath, int vertPos)
        {
            scrollbarPath.Reset();
            scrollbarPath.AddLines(new Point[] { new Point(0, vertPos - 11), new Point(0, vertPos + 11), new Point(scrollbar.Width - 1, vertPos) });
            scrollbarPath.CloseFigure();
        }


        void DrawBufferSummary()
        {
            float increment = (float)data.Size / allDataPathA.Length;
            float dx = 0;
            float x = 0;
            RectangleF rect = new RectangleF();

            allData.BG.Graphics.Clear(allData.BackColor);   // clear the graphics object

            if (!running)
            {
                dx = graphHorizontalSize / increment;   // scale the width of the graph panel for drawing in the allData panel
                x = mostRecentlyDrawnBufferPosition / increment;    // scale the position of the most recent data for drawing in the allData panel
                rect = new RectangleF(x - dx - 1, 0, dx - 1, 256);  // the -1's let the the border of the rectangle show up at the extremes
                allData.BG.Graphics.FillRectangle(Brushes.MidnightBlue, rect);
            }

            data.GetData(0, increment, allDataPathA, allDataPathB);
            GraphicsPath path = new GraphicsPath();
            path.AddLines(allDataPathB);
            Pen penB = new Pen(channelB.PenColor, 2);
            allData.BG.Graphics.DrawPath(penB, path);
            penB.Dispose();
            path.Reset();
            path.AddLines(allDataPathA);
            Pen penA = new Pen(channelA.PenColor, 2);
            allData.BG.Graphics.DrawPath(penA, path);
            penA.Dispose();

            if (!running)
            {
                Pen pen = new Pen(Color.White);
                pen.DashStyle = DashStyle.Dot;
                allData.BG.Graphics.DrawRectangle(pen, rect.X, rect.Y, rect.Width, rect.Height);
                pen.Dispose();
            }

            allData.Invalidate();       // make the panel redraw itself
        }


        void DrawGridLines()
        {
            graph.BG.Graphics.Clear(graph.BackColor);   // clear the graphics object

            Pen gridPen = new Pen(Color.Gray);
            gridPen.DashStyle = DashStyle.Dot;

            float y = (float)countsPerVolt;
            for (int i = 1; i <= numHorizDivisions - 1; i++)
            {
                float x = i * (float)graph.ClientSize.Width / numHorizDivisions;
                graph.BG.Graphics.DrawLine(gridPen, x, 0, x, numVertDivisions * y);
            }
            for (int i = 0; i <= numVertDivisions; i++)
            {
                graph.BG.Graphics.DrawLine(gridPen, 0, y * i, graph.ClientSize.Width, y * i);
            }

            if (scrollbarAIsActive)
                graph.BG.Graphics.DrawLine(Pens.White, 0, vertPosA, graph.ClientSize.Width, vertPosA);
            if (scrollbarBIsActive)
                graph.BG.Graphics.DrawLine(Pens.White, 0, vertPosB, graph.ClientSize.Width, vertPosB);
        }


        private string VoltageString(int count)
        {
            return Math.Round(count / countsPerVolt, 3).ToString("F3") + " V";
        }


        private string TimeString(int idx)
        {
            float time = idx - data.Size - 1;
            if (allAnalogModeRB.Checked)
                time = time / 10000;
            else
                time = time / 20000;
            return Math.Round(time, 3).ToString("F3") + " s";
        }


        private void UpdateTimer_Tick(object sender, EventArgs e)
        {
            if (sloscope == null)
            {
                return;
            }

            int readingsCount = 0;

            // read any existing data from the SLO-scope
            if (running)
            {
                try
                {
                    readingsCount = data.ReadExisting();
                }
                catch (Exception exception)
                {
                    // If there is an error reading the data, then display it and quit.
                    sloscope = null;
                    displayException(exception);
                    Close();
                    return;
                }
                pointsSinceUpdate += readingsCount;
                // these quantities stop being meaningful once they exceed the buffer size, so cap them there
                if (pointsSinceUpdate >= data.Size)
                    pointsSinceUpdate = data.Size - 1;
                if (readingsCount >= data.Size)
                    readingsCount = data.Size - 1;
                mostRecentlyDrawnBufferPosition -= readingsCount;
                MostRecentlyDrawnBoundaryCheck();
            }
            
            DrawBufferSummary();

            if (InStreamingMode() && running)    // stream the data if period > streaming threshold
            {
                if (readingsCount != 0)
                {
                    pointsSinceUpdate = 0;
                    // the array never needs to be bigger than the points that will fill the graph panel
                    //  (usually it can be much smaller)
                    Point[] pointsA = new Point[(int)Math.Min(readingsCount, graphHorizontalSize)];
                    Point[] pointsB = new Point[pointsA.Length];
                    data.GetData(data.Size - 1 - pointsA.Length, 1, pointsA, pointsB);  // we have to read the data in this direction
                    channelA.UpdateScrollingData(pointsA);
                    channelB.UpdateScrollingData(pointsB);
                    mostRecentlyDrawnBufferPosition = data.Size - 1;    // most recently drawn data is most recently received data
                }
            }
            else  // draw the entire screen in a single update if we have enough points
            {
                bool update = false;

                if (!running)
                {
                    // *** this case is the one carried out every timer tick when we aren't running ***
                    data.GetData(mostRecentlyDrawnBufferPosition, -1, fullScreenA, fullScreenB);
                    update = true;                    
                }
                else
                {
                    if (triggerMode.SelectedIndex != 0)
                    {
                        int idx = data.GetTriggerIdx(data.Size - graphHorizontalSize / 2 - 1, readingsCount, 
                            (int)Math.Round((double)triggerLevel.Value * countsPerVolt / 1000), 
                            triggerMode.SelectedIndex == 1, triggerA.Checked);
                        // we want to draw the data with this index in the middle of the graph panel, and we have 
                        // to make sure that all the points we want to draw are new (i.e. have not been drawn)
                        if (idx - graphHorizontalSize / 2 >= data.Size - pointsSinceUpdate)
                        {
                            idx -= graphHorizontalSize / 2;
                            data.GetData(idx, 1, fullScreenA, fullScreenB);
                            pointsSinceUpdate = 0;      // totally clearing this value here might be excessive...
                            mostRecentlyDrawnBufferPosition = idx + graphHorizontalSize - 1;
                            update = true;
                        }
                    }
                    else if (pointsSinceUpdate >= graphHorizontalSize)
                    {
                        data.GetData(data.Size - 1, -1, fullScreenA, fullScreenB);
                        pointsSinceUpdate = 0;
                        mostRecentlyDrawnBufferPosition = data.Size - 1;        // most recently drawn data is most recently received data
                        update = true;
                    }
                }

                if (update)
                {
                    channelA.UpdateFullScreenData(fullScreenA, persistenceCB.Checked && running);
                    channelB.UpdateFullScreenData(fullScreenB, persistenceCB.Checked && running);
                }
                else
                {
                    // we call this to update the persistence lines even if we aren't ready to draw the next data set yet
                    channelA.UpdateFullScreenData(null, persistenceCB.Checked && running);
                    channelB.UpdateFullScreenData(null, persistenceCB.Checked && running);
                }
            }

            // draw grid lines, then draw B data, then draw A data (for order of drawing)
            DrawGridLines();        // draw the grid lines in the graph panel
            DrawTriggerLevel();     // draw a line in the trigger panel (and the graph panel if the trigger level is being dragged)
            channelB.Draw(graph.BG.Graphics, vertPosB - 1);     // draw the data for channel B in the graph panel
            channelA.Draw(graph.BG.Graphics, vertPosA - 1);     // draw the data for channel A in the graph panel
            DrawAnalysisPoint();    // draws the analysis point in the graph panel if scope is stopped and cursor is over graph panel
            graph.Invalidate();     // make the graph panel redraw itself

            DrawScrollbars();       // draw the scroll bars in the scroll bar panel
            UpdateMaxAndMin();      // update max and min labels for channels A and B
        }


        private void DrawAnalysisPoint()
        {
            if (analysisPoint.X < 0 || graphDataPanIsActive)
                return;     // we don't want to draw analysis points under these conditions

            Point[] temp = new Point[1] { analysisPoint };
            Matrix m = channelA.Transform.Clone();
            m.Invert();
            m.TransformPoints(temp);    // transform the horizontal pixel coordinate of the mouse into data buffer space
            int idx = mostRecentlyDrawnBufferPosition - (graphHorizontalSize - temp[0].X);
            Sloscope.Reading reading = data.GetData(idx);

            // draw partially transparent white X's where a vertical line through the cursor intersects channel A and channel B data
            Pen pen = new Pen(Color.FromArgb(180, Color.White), 3);
            float y = (float)reading.channelB / (float)voltsPerDivB.Value + vertPosB;
            graph.BG.Graphics.DrawLine(pen, analysisPoint.X - 5, y - 5, analysisPoint.X + 5, y + 5);
            graph.BG.Graphics.DrawLine(pen, analysisPoint.X - 5, y + 5, analysisPoint.X + 5, y - 5);
            y = (float)reading.channelA / (float)voltsPerDivA.Value + vertPosA;
            graph.BG.Graphics.DrawLine(pen, analysisPoint.X - 5, y - 5, analysisPoint.X + 5, y + 5);
            graph.BG.Graphics.DrawLine(pen, analysisPoint.X - 5, y + 5, analysisPoint.X + 5, y - 5);

            // draw strings for channel A and channel B voltages of the X's and for time coordinate (most recent data is at 0s)
            graph.BG.Graphics.ResetTransform();         // remove graphics object transformation so we can draw coordinate strings
            string timeStr = "T: " + TimeString(idx);   // this will be the longest of the three strings we are about to draw
            SizeF strSize = graph.BG.Graphics.MeasureString(timeStr, SystemFonts.DefaultFont);
            PointF lowerRightCorner = new PointF(strSize.Width + 5, (strSize.Height + 5) * 3 + 5);
            LinearGradientBrush brush = new LinearGradientBrush(new PointF(0, 0), lowerRightCorner, 
                Color.FromArgb(155, Color.MidnightBlue), Color.FromArgb(155, Color.LightBlue));
            graph.BG.Graphics.FillRectangle(brush, 0, 0, lowerRightCorner.X, lowerRightCorner.Y);
            graph.BG.Graphics.DrawString("A: " + VoltageString(reading.channelA), SystemFonts.DefaultFont, Brushes.White, 5, 5);
            graph.BG.Graphics.DrawString("B: " + VoltageString(reading.channelB), SystemFonts.DefaultFont, Brushes.White, 5, 5 + strSize.Height + 5);
            graph.BG.Graphics.DrawString(timeStr, SystemFonts.DefaultFont, Brushes.White, 5, 5 + (strSize.Height + 5) * 2);
            ApplyVerticalGraphTransform(graph);
        }


        private void DrawScrollbars()
        {
            UpdateScrollbar(scrollbarAPath, vertPosA);
            UpdateScrollbar(scrollbarBPath, vertPosB);
            PointF[] points = new PointF[2] { new PointF(0, vertPosA), new PointF(0, vertPosB) };
            scrollbar.BG.Graphics.Transform.TransformPoints(points);
            scrollbar.BG.Graphics.Clear(scrollbar.BackColor);
            scrollbar.BG.Graphics.FillPath(new SolidBrush(channelB.PenColor), scrollbarBPath);
            scrollbar.BG.Graphics.DrawPath(Pens.Black, scrollbarBPath);
            scrollbar.BG.Graphics.ResetTransform();
            scrollbar.BG.Graphics.DrawString("B", SystemFonts.SmallCaptionFont, Brushes.Black, new PointF(0, points[1].Y - 7));

            ApplyVerticalGraphTransform(scrollbar);
            scrollbar.BG.Graphics.FillPath(new SolidBrush(channelA.PenColor), scrollbarAPath);
            scrollbar.BG.Graphics.DrawPath(Pens.Black, scrollbarAPath);
            scrollbar.BG.Graphics.ResetTransform();
            scrollbar.BG.Graphics.DrawString("A", SystemFonts.SmallCaptionFont, Brushes.Black, new PointF(0, points[0].Y - 7));
            ApplyVerticalGraphTransform(scrollbar);
            scrollbar.Invalidate();     // make the scrollbar panel redraw itself
        }


        private void DrawTriggerLevel()
        {
            trigger.BG.Graphics.Clear(trigger.BackColor);

            Pen pen;
            float y = (float)countsPerVolt * (float)triggerLevel.Value / 1000;

            if (triggerA.Checked)
            {
                pen = new Pen(channelA.PenColor, 3);
                y = y / (float)voltsPerDivA.Value + vertPosA;
            }
            else
            {
                pen = new Pen(channelB.PenColor, 3);
                y = y / (float)voltsPerDivB.Value + vertPosB;
            }

            Pen outline = new Pen(Color.Black, 5);
            trigger.BG.Graphics.DrawLine(outline, 0, y, trigger.ClientSize.Width - 2, y);
            trigger.BG.Graphics.DrawLine(pen, 0, y, trigger.ClientSize.Width - 2, y);
            outline.Dispose();
            pen.Dispose();
            trigger.Invalidate();

            if (triggerAdjustIsActive)
            {
                pen = new Pen(Color.Green, 3);
                pen.DashStyle = DashStyle.Dot;
                graph.BG.Graphics.DrawLine(pen, 0, y, graph.ClientSize.Width, y);
                pen.Dispose();
            }
        }


        private void UpdateMaxAndMin()
        {
            int[] maxAndMin = channelB.MaxYAndMinY;
            if (maxAndMin[0] < 0)
            {
                maxB.Visible = false;
                minB.Visible = false;
            }
            else
            {
                maxB.Text = "Max B: " + VoltageString(maxAndMin[0]);
                minB.Text = "Min B: " + VoltageString(maxAndMin[1]);
                maxB.Visible = true;
                minB.Visible = true;
            }
            maxAndMin = channelA.MaxYAndMinY;
            if (maxAndMin[0] < 0)
            {
                maxA.Visible = false;
                minA.Visible = false;
            }
            else
            {
                maxA.Text = "Max A: " + VoltageString(maxAndMin[0]);
                minA.Text = "Min A: " + VoltageString(maxAndMin[1]);
                maxA.Visible = true;
                minA.Visible = true;
            }
        }


        private void periodSelection_ValueChanged(object sender, EventArgs e)
        {
            SetGraphHorizontalSize();
            persistenceCB.Enabled = !InStreamingMode();
            persistenceTime.Enabled = persistenceCB.Enabled;
        }


        private void startStopButton_Click(object sender, EventArgs e)
        {
            try
            {
                if (!running)
                {
                    running = true;     // by first setting running to true, we get a different result from the reset method
                    if (reset)
                        Reset();
                    startStopButton.Text = "Stop";

                    if (allAnalogModeRB.Checked)
                    {
                        sloscope.state = Sloscope.State.TwoAnalog;
                    }
                    else
                    {
                        sloscope.state = Sloscope.State.OneAnalogOneDigital;
                    }

                    SetGraphHorizontalSize();
                    updateTimer.Start();
                    allDataPanIsActive = false;
                    saveButton.Enabled = false;
                    IOA.Enabled = true;
                    IOB.Enabled = true;
                    pointsSinceUpdate = 0;
                }
                else
                {
                    running = false;
                    startStopButton.Text = "Run";
                    saveButton.Enabled = true;
                    IOA.SelectedIndex = 0;      // this will uncheck the output-high checkbox and disable it
                    IOA.Enabled = false;
                    IOB.SelectedIndex = 0;      // this will uncheck the output-high checkbox and disable it
                    IOB.Enabled = false;
                    sloscope.state = Sloscope.State.Off;
                }
            }
            catch (Exception exception)
            {
                // If there is an error, display an error message box and close the program.
                displayException(exception);
                Close();
            }
        }

        private void voltsPerDivA_ValueChanged(object sender, EventArgs e)
        {
            channelA.VoltsPerDivision = (float)voltsPerDivA.Value;
            if (InStreamingMode())
            {
                channelA.Reset();
            }
        }

        private void voltsPerDivB_ValueChanged(object sender, EventArgs e)
        {
            channelB.VoltsPerDivision = (float)voltsPerDivB.Value;
            if (InStreamingMode())
            {
                channelB.Reset();
            }
        }

        private void scrollbar_MouseDown(object sender, MouseEventArgs e)
        {
            Point[] points = new Point[] { e.Location };
            Matrix matrix = scrollbar.BG.Graphics.Transform;
            matrix.Invert();
            matrix.TransformPoints(points);
            if (scrollbarAPath.GetBounds().Contains(points[0]))
            {
                scrollbarAIsActive = true;
            }
            else if (scrollbarBPath.GetBounds().Contains(points[0]))
            {
                scrollbarBIsActive = true;
            }
        }

        private void scrollbar_MouseMove(object sender, MouseEventArgs e)
        {
            if (!scrollbarAIsActive && !scrollbarBIsActive)
                return;
            Point pt = e.Location;
            if (pt.Y < topGraphMargin - 1)
                pt.Y = topGraphMargin - 1;
            else if (pt.Y > scrollbar.ClientSize.Height - bottomGraphMargin + 1)
                pt.Y = scrollbar.ClientSize.Height - bottomGraphMargin + 1;
            Point[] points = new Point[] {pt};
            Matrix matrix = scrollbar.BG.Graphics.Transform;
            matrix.Invert();
            matrix.TransformPoints(points);
            if (scrollbarAIsActive)
                vertPosA = points[0].Y;
            else
                vertPosB = points[0].Y;
        }

        private void scrollbar_MouseUp(object sender, MouseEventArgs e)
        {
            scrollbarAIsActive = false;
            scrollbarBIsActive = false;
        }

        private void allAnalogModeRB_CheckedChanged(object sender, EventArgs e)
        {
            if (running)
            {
                running = false;
                startStopButton_Click(null, null);
            }
        }

        private void MostRecentlyDrawnBoundaryCheck()
        {
            if (mostRecentlyDrawnBufferPosition >= data.Size)
                mostRecentlyDrawnBufferPosition = data.Size - 1;
            if (mostRecentlyDrawnBufferPosition <= graphHorizontalSize)
                mostRecentlyDrawnBufferPosition = graphHorizontalSize + 1;
        }

        private void updateMostRecentlyDrawnBufferPosition(int x)
        {
            mostRecentlyDrawnBufferPosition = (int)(data.Size * (float)x / allData.ClientSize.Width + graphHorizontalSize / 2);
            MostRecentlyDrawnBoundaryCheck();
        }

        private void allData_MouseDown(object sender, MouseEventArgs e)
        {
            if (running)
                return;
            allDataPanIsActive = true;
            updateMostRecentlyDrawnBufferPosition(e.X);
        }

        private void allData_MouseMove(object sender, MouseEventArgs e)
        {
            if (running || !allDataPanIsActive)
                return;
            updateMostRecentlyDrawnBufferPosition(e.X);
        }

        private void allData_MouseUp(object sender, MouseEventArgs e)
        {
            if (running)
                return;
            allDataPanIsActive = false;
        }

        private void graph_MouseDown(object sender, MouseEventArgs e)
        {
            if (running)
                return;
            if (e.Button == MouseButtons.Right)
                graphDataAnalysisActive = !graphDataAnalysisActive;
            else if (e.Button == MouseButtons.Left)
            {
                graphDataPanIsActive = true;
                graphDataPanStartX = e.X;
                this.Cursor = Cursors.Hand;
            }
        }

        private void graph_MouseMove(object sender, MouseEventArgs e)
        {
            if (running)
                return;
            if (graphDataPanIsActive)
            {
                int translation = (graphDataPanStartX - e.X) * graphHorizontalSize / graph.ClientSize.Width;
                if (translation != 0)       // without this check, it is hard to slowly pan when the time domain is small
                    graphDataPanStartX = e.X;
                mostRecentlyDrawnBufferPosition += translation;
                MostRecentlyDrawnBoundaryCheck();
            }
            else if (graphDataAnalysisActive)
            {
                analysisPoint = e.Location;
            }
        }

        private void graph_MouseUp(object sender, MouseEventArgs e)
        {
            if (running)
                return;
            graphDataPanIsActive = false;
            this.Cursor = Cursors.Default;
        }

        private void persistenceTime_ValueChanged(object sender, EventArgs e)
        {
            channelA.NumPens = (int)persistenceTime.Value;
            channelB.NumPens = (int)persistenceTime.Value;
        }

        private void saveButton_Click(object sender, EventArgs e)
        {
            Bitmap bmp = new Bitmap(this.Width, this.Height);
            Graphics g = Graphics.FromImage(bmp);
            g.CopyFromScreen(this.Location, new Point(0, 0), this.Size);
            g.DrawString(DateTime.Now.ToLongDateString() + " " + DateTime.Now.ToLongTimeString(), SystemFonts.CaptionFont, SystemBrushes.ActiveCaptionText, 250, 8);
            g.Dispose();
            if (saveFileDialog.ShowDialog() == DialogResult.OK)
                bmp.Save(saveFileDialog.FileName, System.Drawing.Imaging.ImageFormat.Png);
        }

        private void colorA_MouseClick(object sender, MouseEventArgs e)
        {
            colorDialog.Color = colorA.BackColor;
            if (colorDialog.ShowDialog() == DialogResult.OK)
            {
                colorA.BackColor = colorDialog.Color;
                channelA.PenColor = colorDialog.Color;
                channelA.Reset();
            }
        }

        private void colorB_MouseClick(object sender, MouseEventArgs e)
        {
            colorDialog.Color = colorB.BackColor;
            if (colorDialog.ShowDialog() == DialogResult.OK)
            {
                colorB.BackColor = colorDialog.Color;
                channelB.PenColor = colorDialog.Color;
                channelB.Reset();
            }
        }

        private void IOA_SelectedIndexChanged(object sender, EventArgs e)
        {
            try
            {
                if (((ComboBox)sender).SelectedIndex == 0)
                {
                    sloscope.channelAOutputState = Sloscope.OutputState.Off;
                    outputStateHighA.Checked = false;
                    outputStateHighA.Enabled = false;
                }
                else
                {
                    bool result =
                        warningWithCancel("Warning: You are about to make pin A an output.  Please make sure\r\n" +
                                        "you are not currently using pin A as an input!", "Pin A Output Warning");
                    if (result)
                    {
                        outputStateHighA.Enabled = true;
                        if (outputStateHighA.Checked)
                            sloscope.channelAOutputState = Sloscope.OutputState.High;
                        else
                            sloscope.channelAOutputState = Sloscope.OutputState.Low;
                    }
                    else
                        IOA.SelectedIndex = 0;
                }
            }
            catch (Exception exception)
            {
                displayException(exception);
            }
        }

        private void IOB_SelectedIndexChanged(object sender, EventArgs e)
        {
            try
            {
                if (((ComboBox)sender).SelectedIndex == 0)
                {
                    sloscope.channelBOutputState = Sloscope.OutputState.Off;
                    outputStateHighB.Checked = false;
                    outputStateHighB.Enabled = false;
                }
                else
                {
                    bool result =
                        warningWithCancel("Warning: You are about to make pin A an output.  Please make sure\r\n" +
                                        "you are not currently using pin A as an input!", "Pin A Output Warning");
                    if (result)
                    {
                        outputStateHighB.Enabled = true;
                        if (outputStateHighB.Checked)
                            sloscope.channelBOutputState = Sloscope.OutputState.High;
                        else
                            sloscope.channelBOutputState = Sloscope.OutputState.Low;
                    }
                    else
                        IOB.SelectedIndex = 0;
                }
            }
            catch (Exception exception)
            {
                displayException(exception);
            }
        }

        private void outputStateHighA_CheckedChanged(object sender, EventArgs e)
        {
            if (!running || IOA.SelectedIndex == 0)
                return;
            try
            {
                if (outputStateHighA.Checked)
                    sloscope.channelAOutputState = Sloscope.OutputState.High;
                else
                    sloscope.channelAOutputState = Sloscope.OutputState.Low;
            }
            catch(Exception exception)
            {
                displayException(exception);
            }
        }

        private void outputStateHighB_CheckedChanged(object sender, EventArgs e)
        {
            if (!running || IOB.SelectedIndex == 0)
                return;
            try
            {
                if (outputStateHighB.Checked)
                    sloscope.channelBOutputState = Sloscope.OutputState.High;
                else
                    sloscope.channelBOutputState = Sloscope.OutputState.Low;
            }
            catch (Exception exception)
            {
                displayException(exception);
            }
        }

        private void graph_MouseEnter(object sender, EventArgs e)
        {
            if (!running)
                graphDataAnalysisActive = true;
        }

        private void graph_MouseLeave(object sender, EventArgs e)
        {
            if (!running && graphDataAnalysisActive)
            {
                graphDataAnalysisActive = false;
                analysisPoint = new Point(-1, -1);
            }
        }


        private void UpdateTriggerLevel(Point screenPt)
        {
            Point[] points = new Point[] { screenPt };
            Matrix matrix = trigger.BG.Graphics.Transform;
            matrix.Invert();
            matrix.TransformPoints(points);
            decimal millivolts;
            if (triggerA.Checked)
                 millivolts = (decimal)Math.Round((points[0].Y - vertPosA) * 1000 * (double)voltsPerDivA.Value / countsPerVolt);
            else
                millivolts = (decimal)Math.Round((points[0].Y - vertPosB) * 1000 * (double)voltsPerDivB.Value / countsPerVolt);
            if (millivolts > triggerLevel.Maximum)
                millivolts = triggerLevel.Maximum;
            if (millivolts < triggerLevel.Minimum)
                millivolts = triggerLevel.Minimum;
            triggerLevel.Value = millivolts;
        }

        private void trigger_MouseDown(object sender, MouseEventArgs e)
        {
            triggerAdjustIsActive = true;
            UpdateTriggerLevel(e.Location);
        }


        private void trigger_MouseMove(object sender, MouseEventArgs e)
        {
            if (!triggerAdjustIsActive)
                return;
            UpdateTriggerLevel(e.Location);
        }

        private void trigger_MouseUp(object sender, MouseEventArgs e)
        {
            triggerAdjustIsActive = false;
        }

        public static String convertExceptionToString(Exception exception)
        {
            StringBuilder stringBuilder = new StringBuilder();

            do
            {
                stringBuilder.Append(exception.Message + "  ");

                if (exception is Win32Exception)
                {
                    stringBuilder.Append("Error code 0x" + ((Win32Exception)exception).NativeErrorCode.ToString("x") + ".  ");
                }

                exception = exception.InnerException;
            }
            while (exception != null);

            return stringBuilder.ToString().Trim();
        }

        public void displayException(String broadErrorSentence)
        {
            displayException(new Exception(broadErrorSentence));
        }

        public void displayException(Exception exception, String broadErrorSentence)
        {
            displayException(new Exception(broadErrorSentence, exception));
        }

        public void displayException(Exception exception)
        {
            String message = convertExceptionToString(exception);
            MessageBox.Show(message, this.Text, MessageBoxButtons.OK, MessageBoxIcon.Error);
        }

        /// <summary>
        /// Warn the user about something.
        /// </summary>
        /// <param name="caption">Should be a noun phrase, like Firmware Upgrade.</param>
        public Boolean warningWithCancel(string text, string caption)
        {
            return DialogResult.OK == MessageBox.Show(text, this.Text + " " + caption, MessageBoxButtons.OKCancel, MessageBoxIcon.Warning);
        }

        /// <summary>
        /// Asks the user a OK/Cancel question using a dialog box.
        /// </summary>
        /// <param name="caption">Should be a noun phrase, like Firmware Upgrade.</param>
        /// <returns></returns>
        public Boolean confirm(string question, string caption)
        {
            return DialogResult.OK == MessageBox.Show(question, this.Text + " " + caption,
                MessageBoxButtons.OKCancel, MessageBoxIcon.Warning, MessageBoxDefaultButton.Button2);
        }
    }



    public class CustomPanel : Panel
    {
        public BufferedGraphics BG = null;

        public CustomPanel()
        {

        }

        ~CustomPanel()
        {
            this.Dispose();
        }

        public void CreateBufferedGraphics()
        {
            Graphics g = this.CreateGraphics();
            BG = BufferedGraphicsManager.Current.Allocate(g, this.ClientRectangle);
            BG.Graphics.Clear(this.BackColor);
            g.Dispose();
        }

        protected override void Dispose(bool disposing)
        {
            if (BG != null)
                BG.Dispose();
            base.Dispose(disposing);
        }

        
        protected override void OnPaint(PaintEventArgs e)
        {
            if (BG != null)
                BG.Render(e.Graphics);
            else
                base.OnPaint(e);
        }

        protected override void OnPaintBackground(PaintEventArgs e)
        {
            if (BG == null)
                base.OnPaintBackground(e);
        }
    }
}
