// This file defines the API for using the sloscope.

#ifndef SLOSCOPE_H_DLL_MODE
  #define SLOSCOPE_H_DLL_MODE
#endif

struct sloscope;

#ifdef __cplusplus
extern "C" {
#endif

/*  sloscopeConnectToFirst: The easiest way to connect to an sloscope.
	Call sloscopeFree() when you are done using this sloscope object.
    If some other program is already using the sloscope, then this fails.
 */
SLOSCOPE_H_DLL_MODE struct sloscope * __stdcall sloscopeConnect();

/*  Gets the serial number of the specified SLO-scope in the list.
    listHandle should be the handle returned by sloscopeListCreate().
 */
SLOSCOPE_H_DLL_MODE wchar_t * __stdcall sloscopeGetSerialNumber(struct sloscope *);

/*  Creates a list of all the SLO-scopes connected to the computer.
    Returns a handle to the list that should be used as the first argument
	to all the sloscopeList* functions.
 */
SLOSCOPE_H_DLL_MODE void * __stdcall sloscopeListCreate();

/*  Gets the serial number of the specified SLO-scope in the list.
    listHandle should be the handle returned by sloscopeListCreate().
 */
SLOSCOPE_H_DLL_MODE wchar_t * __stdcall sloscopeListGetSerialNumber(void * listHandle, unsigned char index);

/*  Connects to the specified sloscope in the list.
	Call sloscopeFree() when you are done using this sloscope object.
    listHandle should be the handle returned by sloscopeListCreate().
 */
SLOSCOPE_H_DLL_MODE struct sloscope * __stdcall sloscopeListConnect(void * listHandle, unsigned char index);

/*  Destroys the list of sloscopes connected to the computer.
	Call this function when you done using the list so you can free up that memory.
    listHandle should be the handle returned by sloscopeListCreate().
 */
SLOSCOPE_H_DLL_MODE void __stdcall sloscopeListDestroy(void * listHandle);

/*  sloscopeFree: De-allocates all memory used by the sloscope object,
    and frees it so the device can be used by other programs.
 */
SLOSCOPE_H_DLL_MODE void __stdcall sloscopeFree(struct sloscope *);

/*  sloscopeFlushBuffers: flushes the buffers, so that you are guaranteed to not
    get old data at the beginning when you next read oscilloscope data.
 */
SLOSCOPE_H_DLL_MODE void __stdcall sloscopeFlushBuffers(struct sloscope *);

/*  sloscopeSetState and sloscopeGetState.
    This function lets you set and get the state of the sloscope.  The states available are:
		SLOSCOPE_STATE_OFF
			The sloscope will be off, and not taking any measurements.

		SLOSCOPE_STATE_2ANALOG
			The sloscope will take 10kHz 8-bit analog measurements on Channels A and B. 

		SLOSCOPE_STATE_1ANALOG_1DIGITAL
			The sloscope will take 20kHz 7-bit analog measurement of Channel A, and
			20kHz 1-bit digital measurement of Channel B.
 */
SLOSCOPE_H_DLL_MODE void __stdcall sloscopeSetState(struct sloscope *, unsigned char state);
SLOSCOPE_H_DLL_MODE unsigned char __stdcall sloscopeGetState(struct sloscope *);

/*  sloscopeRead:
    Reads data from the sloscope.

	Precondition:
		Before calling this function, you must either successfully call sloscopeSetState with
		a state other than SLOSCOPE_STATE_OFF,  OR you must call sloscopeGetState and it must
		successfully return a state other than

	Arguments:
		buffer: A pointer to the buffer to read the data in to.
        bufferLength: The length of the buffer.  sloscopeRead will never write more than
	        that many bytes in to the buffer.
	    readMode: This dictates the behavior of this functions.  It should be one of the following:

			READ_MODE_FILL_BUFFER: Read data until the buffer has been filled.
				(Note: If bufferLength is not a multiple of 2, then we can not fill the buffer
				because the data comes in 2-byte chunks.  In that case, sloscopeRead
				will fill the buffer until less than 2 bytes is left.)

			READ_MODE_EXISTING: Read all the data that has been received by the computer
			  and then return.  If this mode is used, then sloscopeRead will take a very short
			  time to return because it will not spend any time waiting for data.
			 
		timeoutMS (optional):
			If readMode is READ_MODE_EXISTING, then timeoutMS is ignored.
			If readMode is READ_MODE_FILL_BUFFER and timeoutMS is non-zero, then
			the function is guaranteed to return within timeoutMS milliseconds.

    Return Value:
	    The number of bytes that were read in to the buffer.
	    This can be zero, and it will always be a multiple of 2.

    Data format:
        The data comes in 2-byte chunks.
		The format depends on the oscilloscope state.

		If the state is OSCILLOSCOPE_STATE_2ANALOG, then the format is:
			First Byte: Channel A analog value, 0 is GND, 0xFF is VCC (about 4.7V).
			Second Byte: Channel B analog value, 0 is GND, 0xFF is VCC (nominally 4.7V).
			Each reading represents 0.1 milliseconds of time.

		If the state is OSCILLOSCOPE_STATE_1ANALOG_1DIGITAL, then the format is:
		    First Byte: Channel A 7-bit analog value, 0 is GND, 0xFE is VBUS (nominally 5V).  The low bit is always 0.
			Second Byte: Channel B digital value, 0 is Low, 0xFF is High.
			Each reading represents 0.05 millieconds of time.
 */
SLOSCOPE_H_DLL_MODE unsigned __int32 __stdcall sloscopeRead(struct sloscope *, unsigned char * buffer, unsigned __int32 bufferLength, unsigned char readMode, unsigned __int32 timeoutMS);
#define READ_MODE_FILL_BUFFER 0
#define READ_MODE_EXISTING 1

/*  sloscopeCheckMissedReadings()
    This function returns the missed readings counter.  This is the number
	of readings that the device took, but was unable to send to the computer
	because the device was taking readings faster than the USB/computer
	could read them.

	sloscopeCheckMissedFrames()
	This function returns the missed frames counter.  This is the number of USB
	frames (1 millisecond time periods) that we did not receive a packet from,
	either because the device was too slow at taking measurements (in which
	case we would expect missedReadings to be zero), or the computer was too
	slow at asking for them (in which case we would expect the missedReadings
	to be non-zero).

	Both of these counters are reset when the sloscope object is created, 
	when the counter is checked using the functions below, and also whenever
	any of these functions is called:
		sloscopeFlushBuffers()
		sloscopeSetState()
 */
SLOSCOPE_H_DLL_MODE unsigned __int32 __stdcall sloscopeCheckMissedReadings(struct sloscope *);
SLOSCOPE_H_DLL_MODE unsigned __int32 __stdcall sloscopeCheckMissedFrames(struct sloscope *);

/*  sloscopeSetOutputState and sloscopeGetOutputState
    These functions allow you to use Channels A and B as outputs while the sloscope is on.
    
	channelAOutputState and channelBOutputState should both be one of the following:

		SLOSCOPE_OUTPUT_OFF
			The channel will be in a high impedance state so it will not be outputing anything.
		
		SLOSCOPE_OUTPUT_LOW
			The channel will drive low (0V).
			WARNING: This can cause short circuits if you are not careful.  There are no
			current-limiting resistors on the channel.

		SLOSCOPE_OUTPUT_HIGH 
			The channel will drive high (nominally 5V).
			WARNING: This can cause short circuits if you are not careful.  There are no
			current-limiting resistors on the channel.

		SLOSCOPE_OUTPUT_NO_CHANGE
			The channel's output state will not change: it will keep doing whatever it was
			doing before.

	The return value of sloscopeGetOutputState is channelAOutputState | (channelBOutputState << 8)
 */
SLOSCOPE_H_DLL_MODE void __stdcall sloscopeSetOutputState(struct sloscope *, unsigned char channelAOutputState, unsigned char channelBOutputState);
SLOSCOPE_H_DLL_MODE unsigned __int16 __stdcall sloscopeGetOutputState(struct sloscope *);

/*  sloscopeSetPeriod and sloscopeGetPeriod: Sets and gets the period of the sloscope.
    The main use of this is tuning the slo-scope for maximum performance.
	The slo-scope can be used without calling this function because there is a good default
	period programmed in to the firmware.  But it's nice to have this function just in case
	we ever discover that the default period in the firmware is bad.

	The period of the oscilloscope in microseconds is
      (1/12 us) * (period + 1)
 */
SLOSCOPE_H_DLL_MODE void __stdcall sloscopeSetPeriod(struct sloscope *, unsigned __int16 period);
SLOSCOPE_H_DLL_MODE unsigned __int16 __stdcall sloscopeGetPeriod(struct sloscope *);

#ifdef __cplusplus
}
#endif

/* WINUSB_HELPER ERROR CODES
   Every winusb_helper function will call SetLastError(0) on success.
   So if GetLastError() returns a non-zero value, there was some error.
   Sometimes GetLastError() will be some standard system error, and sometimes
   they will be one of the custom error codes defined below.
 */
#define SLOSCOPE_ERROR_STATE_INVALID         0x21010000
#define SLOSCOPE_ERROR_READ_PIPE             0x21020000
#define SLOSCOPE_ERROR_GET_OVERLAPPED_RESULT 0x21030000